/*
 * Interprocess communication message abstraction.
 *
 * Copyright (C) 2018, 2019, 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <string.h>

#include "cap_alloc.h"
#include "message.h"
#include "mem_ipc.h"
#include "util_ipc.h"



/* Calculate the given length rounded up to whole word increments. */

static size_t get_size_in_words(size_t length)
{
  size_t mask = sizeof(l4_umword_t) - 1;
  return (length + mask) & ~mask;
}



/* Message operations. */

/* Discard capabilities for transfer to other tasks. */

void ipc_message_discard(ipc_message_t *msg)
{
  unsigned int i;

  for (i = 0; i < msg->discarded_items; i++)
    ipc_cap_free_um(msg->to_discard[i]);
}

/* Initialise a message structure with the given number of expected items. */

long ipc_message_expect(ipc_message_t *msg, unsigned int expected_items)
{
  return ipc_message_expect_capabilities(msg, expected_items);
}

/* Free capabilities expected in messages. */

void ipc_message_free(ipc_message_t *msg)
{
  _free_expected_capabilities(&msg->bregs, msg->expected_items);
}

/* Clear message attributes for sending and response handling. */

void ipc_message_new(ipc_message_t *msg)
{
  /* Set a default for expected items. */

  msg->expected_items = 0;
  msg->receive_items = 0;
  ipc_message_reset(msg);
}

/* Open an incoming message by resetting sending state and preserving the
   message registers plus the buffer registers for the expected number of
   items. */

void ipc_message_open(ipc_message_t *msg)
{
  ipc_message_reset(msg);
  ipc_message_preserve_buffer_registers(msg);
  ipc_message_preserve_message_registers(msg);
}

/* Preserve the buffer registers. */

void ipc_message_preserve_buffer_registers(ipc_message_t *msg)
{
  unsigned int i;

  /* Reference the UTCB virtual registers. */

  l4_buf_regs_t *bregs = l4_utcb_br();

  msg->bregs.bdr = bregs->bdr;

  /* NOTE: Support a mixture of expected items and received flexpages. */

  if ((msg->expected_items) || (msg->receive_items))
  {
    for (i = 0; (i < msg->expected_items) || (i < msg->receive_items * 2); i++)
      msg->bregs.br[i] = bregs->br[i];
  }
}

/* Preserve the message registers. */

void ipc_message_preserve_message_registers(ipc_message_t *msg)
{
  /* Compute the number of message register words affected. */

  unsigned int i, affected_words = l4_msgtag_words(msg->tag) +
                                   2 * l4_msgtag_items(msg->tag);

  /* Reference the UTCB virtual registers. */

  l4_msg_regs_t *mregs = l4_utcb_mr();

  /* Copy message register information. */

  for (i = 0; i < affected_words; i++)
    msg->mregs.mr[i] = mregs->mr[i];
}

/* Restore the message and buffer registers. */

void ipc_message_prepare(ipc_message_t *msg)
{
  ipc_message_restore_buffer_registers(msg);
  ipc_message_restore_message_registers(msg);
}

/* Restore buffer registers so that item expectations can be set. */

void ipc_message_restore_buffer_registers(ipc_message_t *msg)
{
  unsigned int i;

  /* Reference the UTCB virtual registers. */

  l4_buf_regs_t *bregs = l4_utcb_br();

  /* Restore buffer register information. */

  bregs->bdr = msg->bregs.bdr;

  /* NOTE: Support a mixture of expected items and received flexpages. */

  if ((msg->expected_items) || (msg->receive_items))
  {
    for (i = 0; (i < msg->expected_items) || (i < msg->receive_items * 2); i++)
      bregs->br[i] = msg->bregs.br[i];
  }
}

/* Restore message registers to communicate recorded data and items. */

void ipc_message_restore_message_registers(ipc_message_t *msg)
{
  /* Compute the number of message register words affected. */

  unsigned int i, affected_words = msg->words + 2 * msg->items;

  /* Reference the UTCB virtual registers. */

  l4_msg_regs_t *mregs = l4_utcb_mr();

  /* Restore message register information. */

  for (i = 0; i < affected_words; i++)
    mregs->mr[i] = msg->mregs.mr[i];
}

/* Prepare and send a reply using the message. */

void ipc_message_reply(ipc_message_t *msg)
{
  ipc_message_prepare(msg);
  msg->tag = util_ipc_reply(ipc_message_reply_tag(msg));
}

/* Prepare and send a request using the message, involving the given operation,
   directed at the indicated endpoint. Open a message for the reply. */

void ipc_message_request(ipc_message_t *msg, int op, l4_cap_idx_t endpoint)
{
  ipc_message_prepare(msg);
  msg->tag = util_ipc_request(ipc_message_request_tag(msg, op), endpoint);
  ipc_message_open(msg);
}

/* Prepare and send a message, involving the given operation, directed at the
   indicated endpoint. No reply is expected. */

void ipc_message_send(ipc_message_t *msg, int op, l4_cap_idx_t endpoint)
{
  ipc_message_prepare(msg);
  msg->tag = util_ipc_send(ipc_message_request_tag(msg, op), endpoint);
}

/* Reset the state of the message for sending purposes. */

void ipc_message_reset(ipc_message_t *msg)
{
  /* Initialise words and items for sending. */

  msg->discarded_items = 0;
  msg->words = 0;
  msg->items = 0;

  /* Message label overriding. */

  msg->new_label = 0;

  /* Server control. */

  msg->terminating = 0;
}

/* Wait for an incoming message. */

void ipc_message_wait(ipc_message_t *msg, l4_umword_t *label)
{
  ipc_message_restore_buffer_registers(msg);
  msg->tag = l4_ipc_wait(l4_utcb(), label, L4_IPC_NEVER);
}



/* Add a capability to the message. */

void ipc_message_add_capability(ipc_message_t *msg, l4_cap_idx_t cap)
{
  ipc_message_export_capability(msg, msg->items++, cap);

  /* NOTE: Might use the "grant" operation instead of explicitly discarding. */

  if (cap & IPC_DISCARD_CAP_FLAG)
    ipc_message_discard_capability(msg, cap & ~IPC_DISCARD_CAP_FLAG);
}

/* Add an item to the message. */

void ipc_message_add_item(ipc_message_t *msg, l4_cap_idx_t cap)
{
  ipc_message_export_capability(msg, msg->items++, cap);

  /* NOTE: Might use the "grant" operation instead of explicitly discarding. */

  if (cap & IPC_DISCARD_CAP_FLAG)
    ipc_message_discard_capability(msg, cap & ~IPC_DISCARD_CAP_FLAG);
}

/* Add a flexpage to the message using a type to combine the "hot spot" and
   flexpage. */

void ipc_message_add_fpage(ipc_message_t *msg, l4_snd_fpage_t fpage)
{
  /* Test for {0, l4_fpage_invalid()} as the flexpage to avoid setting an
     optional flexpage result, useful for the page fault handler. */

  if (fpage.snd_base || fpage.fpage.raw)
    ipc_message_export_fpage(msg, msg->items++, fpage);
}

/* Add a receive window flexpage item to the message. */

void ipc_message_add_receive_fpage(ipc_message_t *msg, l4_snd_fpage_t fpage)
{
  ipc_message_receive_fpage(msg, msg->receive_items++, fpage);
}

/* Add a flexpage to the message. */

void ipc_message_add_page(ipc_message_t *msg, l4_umword_t hot_spot,
                          l4_fpage_t fpage)
{
  ipc_message_export_page(msg, msg->items++, hot_spot, fpage);
}

/* Add a word value to the message. */

void ipc_message_add_word(ipc_message_t *msg, l4_umword_t value)
{
  /* NOTE: Should raise an exception if there are items. */

  if (!msg->items)
    msg->mregs.mr[msg->words++] = value;
}

/* Add the given dimensioned data to the message. */

void ipc_message_add_data(ipc_message_t *msg, const char *value, size_t length)
{
  void *target = ipc_message_reserve_data(msg, length);

  memcpy(target, value, length);
}

/* Add the given null-terminated character string to the message. */

void ipc_message_add_string(ipc_message_t *msg, const char *value)
{
  ipc_message_add_data(msg, value, strlen(value) + 1);
}

/* Propagate the given item to the message. */

void ipc_message_propagate_item(ipc_message_t *msg, l4_cap_idx_t cap)
{
  ipc_message_propagate_capability(msg, msg->items++, cap);
}

/* Indicate an error condition in the message. */

void ipc_message_send_error(ipc_message_t *msg, long error)
{
  msg->new_label = error;
}

/* Return a word value from the message. */

l4_umword_t ipc_message_get_word(ipc_message_t *msg, unsigned int word)
{
  return msg->mregs.mr[word];
}

/* Return a pointer to a word in the message. */

l4_umword_t *ipc_message_get_word_address(ipc_message_t *msg, unsigned int word)
{
  return &msg->mregs.mr[word];
}

/* Return the number of received items. */

unsigned int ipc_message_number_of_items(ipc_message_t *msg)
{
  return l4_msgtag_items(msg->tag);
}

/* Return the number of received words. */

unsigned int ipc_message_number_of_words(ipc_message_t *msg)
{
  return l4_msgtag_words(msg->tag);
}

/* Message tag generation. */

l4_msgtag_t ipc_message_reply_tag(ipc_message_t *msg)
{
  return l4_msgtag(msg->new_label, msg->words, msg->items, 0);
}

l4_msgtag_t ipc_message_request_tag(ipc_message_t *msg, int op)
{
  return l4_msgtag(op, msg->words, msg->items, 0);
}

/* Reserve space for dimensioned data, returning a pointer to use to add the
   data. The length of the data is incorporated into the message. */

void *ipc_message_reserve_data(ipc_message_t *msg, size_t length)
{
  /* Indicate the length of the data. */

  ipc_message_add_word(msg, length);

  /* Reserve the space. */

  return ipc_message_reserve_words(msg, length);
}

/* Reserve space for dimensioned data, returning a pointer to use to add the
   data. The length is not incorporated into the message. */

void *ipc_message_reserve_words(ipc_message_t *msg, size_t length)
{
  /* Calculate the rounded-up size of the data and the necessary padding. */

  size_t size = get_size_in_words(length);
  size_t padding = size - length;
  void *target;

  /* NOTE: Should test the length against the capacity of the message. */

  /* Obtain the location of the next word, where the data will be copied. */

  target = &msg->mregs.mr[msg->words];

  /* Pad the data and update the number of words. */

  memset(target + length, 0, padding);
  msg->words += size / sizeof(l4_umword_t);

  return target;
}



/* Discard a capability after replying. */

void ipc_message_discard_capability(ipc_message_t *msg, l4_cap_idx_t cap)
{
  msg->to_discard[msg->discarded_items++] = cap;
}

/* Discard a dataspace. */

void ipc_message_discard_dataspace(ipc_message_t *msg, l4re_ds_t mem, l4_addr_t addr)
{
  ipc_message_discard_capability(msg, mem);
  ipc_detach_dataspace((void *) addr);
}

/* Reserve the given number slots from zero for incoming capabilities. */

long ipc_message_expect_capabilities(ipc_message_t *msg, int number)
{
  msg->expected_items = number;
  return _expect_capabilities(&msg->bregs, number);
}

/* Reserve a slot for an incoming capability. */

long ipc_message_expect_capability(ipc_message_t *msg, int item)
{
  if (item >= (int) msg->expected_items)
    msg->expected_items = item + 1;

  return _expect_capability(&msg->bregs, item);
}

/* Export a capability at the given position in the message. */

void ipc_message_export_capability(ipc_message_t *msg, int item, l4_cap_idx_t ref)
{
  msg->mregs.mr[msg->words + item * 2] = 0 | L4_ITEM_MAP;
  msg->mregs.mr[msg->words + item * 2 + 1] = l4_obj_fpage(ref, 0, L4_FPAGE_RWX).raw;
}

/* Export a flexpage at the given position in the message. Here, the snd_base
   member of the flexpage structure is used to hold the "hot spot" value. */

void ipc_message_export_fpage(ipc_message_t *msg, int item, l4_snd_fpage_t fpage)
{
  msg->mregs.mr[msg->words + item * 2] = l4_map_control(fpage.snd_base, L4_FPAGE_CACHEABLE, 0);
  msg->mregs.mr[msg->words + item * 2 + 1] = fpage.fpage.raw;
}

/* Export a flexpage at the given position in the message. */

void ipc_message_export_page(ipc_message_t *msg, int item, l4_umword_t hot_spot, l4_fpage_t fpage)
{
  msg->mregs.mr[msg->words + item * 2] = l4_map_control(hot_spot, L4_FPAGE_CACHEABLE, 0);
  msg->mregs.mr[msg->words + item * 2 + 1] = fpage.raw;
}

/* Specify a receive window flexpage item in the message. */

void ipc_message_receive_fpage(ipc_message_t *msg, int item, l4_snd_fpage_t fpage)
{
  msg->bregs.br[item * 2] = l4_map_control(fpage.snd_base, 0, 0);
  msg->bregs.br[item * 2 + 1] = fpage.fpage.raw;
}



/* Import from the message the capability at the given item position, updating
   the buffer registers for future capabilities. */

long ipc_message_import_capability(ipc_message_t *msg, int item, l4_cap_idx_t *ref)
{
  long err;

  err = _import_capability(msg->tag, &msg->bregs, &msg->mregs, item, ref);
  if (err)
    return err;

  return ipc_message_expect_capability(msg, item);
}

/* Import from the message a dataspace, mapping it to an address, updating the
   buffer registers for future capabilities. */

long ipc_message_import_dataspace(ipc_message_t *msg, int item, l4re_ds_t *mem, l4_addr_t *addr)
{
  long err;

  err = _import_dataspace(msg->tag, &msg->bregs, &msg->mregs, item, mem, addr);
  if (err)
    return err;

  return ipc_message_expect_capability(msg, item);
}

/* Import from the message the flexpage at the given item position. */

long ipc_message_import_fpage(ipc_message_t *msg, int item, l4_snd_fpage_t *fpage)
{
  return _import_fpage(msg->tag, &msg->bregs, &msg->mregs, item, fpage);
}

/* Export a capability, discarding it by propagating it to the recipient. */

void ipc_message_propagate_capability(ipc_message_t *msg, int item, l4_cap_idx_t ref)
{
  ipc_message_export_capability(msg, item, ref);
  ipc_message_discard_capability(msg, ref);
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
